/* Support for loading and querying Java 1.5 class file data structure. 
   Should handle all class files, but ignores some attibutes (including all
   non-standard ones).
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <errno.h>
#include "basics.h"
#include "class.h"

#define CF_ALLOC(x) alloc(x)

/* Constant pool utilities. */
static Constant get_constant(Class *class, u2 index) {
  assert (index >= 1 && index <= class->constant_pool_count - 1);
  return class->constant_pool[index-1];
}

u1 get_constant_tag(Class *class, u2 index) {
  return get_constant(class,index).tag;
}

char *get_utf8(Class *class,u2 index) {
  Constant c = get_constant(class,index);
  assert (c.tag == CONSTANT_Utf8);
  return (char *) (c.u.Utf8.bytes);
}

char *get_class(Class *class,u2 index) {
  Constant c = get_constant(class,index);
  assert (c.tag == CONSTANT_Class);
  return get_utf8(class,c.u.Class.name_index);
}

char *get_string(Class *class,u2 index) {
  Constant c = get_constant(class,index);
  assert (c.tag == CONSTANT_String);
  return get_utf8(class,c.u.String.string_index);
}

int32_t get_integer(Class *class,u2 index) {
  Constant c = get_constant(class,index);
  assert (c.tag == CONSTANT_Integer);
  return (int32_t) c.u.Integer.bytes;
}

char *get_name(Class *class,u2 index) {
  Constant c = get_constant(class,index);
  assert (c.tag == CONSTANT_NameAndType);
  return get_utf8(class,c.u.NameAndType.name_index);
}

char *get_type(Class *class,u2 index) {
  Constant c = get_constant(class,index);
  assert (c.tag == CONSTANT_NameAndType);
  return get_utf8(class,c.u.NameAndType.descriptor_index);
}

char *get_field_class(Class *class,u2 index) {
  Constant c = get_constant(class,index);
  assert (c.tag == CONSTANT_Fieldref);
  return get_class(class,c.u.Fieldref.class_index);
}

char *get_field_name(Class *class,u2 index) {
  Constant c = get_constant(class,index);
  assert (c.tag == CONSTANT_Fieldref);
  return get_name(class,c.u.Fieldref.name_and_type_index);
}

char *get_field_type(Class *class,u2 index) {
  Constant c = get_constant(class,index);
  assert (c.tag == CONSTANT_Fieldref);
  return get_type(class,c.u.Fieldref.name_and_type_index);
}

char *get_method_class(Class *class,u2 index) {
  Constant c = get_constant(class,index);
  assert (c.tag == CONSTANT_Methodref);
  return get_class(class,c.u.Methodref.class_index);
}

char *get_method_name(Class *class,u2 index) {
  Constant c = get_constant(class,index);
  assert (c.tag == CONSTANT_Methodref);
  return get_name(class,c.u.Methodref.name_and_type_index);
}

char *get_method_type(Class *class,u2 index) {
  Constant c = get_constant(class,index);
  assert (c.tag == CONSTANT_Methodref);
  return get_type(class,c.u.Methodref.name_and_type_index);
}

char *get_imethod_class(Class *class,u2 index) {
  Constant c = get_constant(class,index);
  assert (c.tag == CONSTANT_InterfaceMethodref);
  return get_class(class,c.u.InterfaceMethodref.class_index);
}

char *get_imethod_name(Class *class,u2 index) {
  Constant c = get_constant(class,index);
  assert (c.tag == CONSTANT_InterfaceMethodref);
  return get_name(class,c.u.InterfaceMethodref.name_and_type_index);
}

char *get_imethod_type(Class *class,u2 index) {
  Constant c = get_constant(class,index);
  assert (c.tag == CONSTANT_InterfaceMethodref);
  return get_type(class,c.u.InterfaceMethodref.name_and_type_index);
}

/* Method table utilities.*/

/* Find a method (by name and type) in a class's method table.
   (Note that we can make no assumptions that a given
   name or type appears only once in the pool, so it is
   not safe to merely compare indices.)
 */
Method *find_method(Class *class,char *method_name, char *method_type) {
  int i = 0;
  Method *mp = class->methods;
  for (; i < class->methods_count; i++, mp++) {
    if (STR_EQ(get_utf8(class,mp->name_index), method_name) &&
	STR_EQ(get_utf8(class,mp->descriptor_index), method_type))
      return mp;
  }
  return 0; /* method not found */
}
  
/* Descriptor utilities. */
/* Consume one parameter descriptor from a sequence of them. */
static char *consume_descriptor(char* s) {
  if (*s == '[')           
    return consume_descriptor(s+1);
  else if (*s == 'L') {    
    for (s++; *s != ';'; s++);
    return s+1;
  } else  /* just assume a valid BaseType code */
    return s+1;
}

/* Return size (in words) of value represented by descriptor. 
   (Only first character of descriptor is needed.) */
size_t size_descriptor(char *s) {
  switch (*s) {
  case 'V':
    return 0;
  case 'D':
  case 'J':
    return 2;
  default:
    return 1;
  }
}

/* Return the number of words occupied by method's arguments,
   given method's type descriptor. */
size_t count_args(char *s) {
  size_t i = 0;
  assert (*s == '(');
  s++;
  while (*s != ')') {
    i += size_descriptor(s);
    s = consume_descriptor(s);
  }
  return i;
}

/* Return the number of words occupied by method's result,
   given method's type descriptor. */
size_t count_result(char *s) {
  while (*(s++) != ')');  /* skip over argument */
  return size_descriptor(s);
}

/* Reading class file into internal data structure. */

/* Supporting utilities. */

static u1 readu1(FILE *cf) {
  u1 u = getc(cf);
  return u;
}

static u2 readu2(FILE *cf) {
  u2 u = getc(cf);
  u = (u << 8) | (getc(cf));
  return u;
}

static u4 readu4(FILE *cf) {
  u4 u = getc(cf);
  u = (u << 8) | (getc(cf));
  u = (u << 8) | (getc(cf));
  u = (u << 8) | (getc(cf));
  return u;
}

/* expects attribute name index has already been read */
static void skip_attribute(FILE *cf) {
  int i;
  u4 length = readu4(cf);
  for (i = 0; i < length; i++) 
    readu1(cf);
}

static void skip_attributes(FILE *cf) {
  int i;
  u2 count = readu2(cf); /* attributes count */
  for (i = 0; i < count; i++) {
    readu2(cf);   /* skip attribute name index */
    skip_attribute(cf);
  }
}

static Constant *read_constant_pool(FILE *cf,int count) {
  int i;
  Constant *constant_pool = (Constant *) CF_ALLOC(sizeof(Constant) * (count - 1));
  Constant *c;
  for (c = constant_pool; c < constant_pool + count - 1; c++) {
    c->tag = readu1(cf);
    switch (c->tag) {
    case CONSTANT_Class:
      c->u.Class.name_index = readu2(cf);
      break;
    case CONSTANT_Fieldref:
      c->u.Fieldref.class_index = readu2(cf); 
      c->u.Fieldref.name_and_type_index = readu2(cf);
      break;
    case CONSTANT_Methodref:
      c->u.Methodref.class_index = readu2(cf);
      c->u.Methodref.name_and_type_index = readu2(cf);
      break;
    case CONSTANT_InterfaceMethodref:
      c->u.InterfaceMethodref.class_index = readu2(cf);
      c->u.InterfaceMethodref.name_and_type_index = readu2(cf);
      break;
    case CONSTANT_String:
      c->u.String.string_index = readu2(cf);
      break;
    case CONSTANT_Integer:
      c->u.Integer.bytes = readu4(cf);
      break;
    case CONSTANT_Float:
      c->u.Float.bytes = readu4(cf);
      break;
    case CONSTANT_Long:
      c->u.Long.high_bytes = readu4(cf);
      c->u.Long.low_bytes = readu4(cf);
      c++; /* skip extra entry */
      break;
    case CONSTANT_Double:
      c->u.Double.high_bytes = readu4(cf);
      c->u.Double.low_bytes = readu4(cf);
      c++; /* skip extra entry */
      break;
    case CONSTANT_NameAndType:
      c->u.NameAndType.name_index = readu2(cf); 
      c->u.NameAndType.descriptor_index = readu2(cf);
      break;
    case CONSTANT_Utf8:
      c->u.Utf8.length = readu2(cf); 
      c->u.Utf8.bytes = (u1 *) (CF_ALLOC(sizeof(u1) * (c->u.Utf8.length + 1)));
      for (i = 0; i < c->u.Utf8.length; i++)
	c->u.Utf8.bytes[i] = readu1(cf);
      c->u.Utf8.bytes[i] = '\0';
      break;
    default:
      die("bad constant pool entry tag %d",c->tag);
    }
  }
  return constant_pool;
}

static Handler *read_handlers(FILE *cf, Class *cl, int count) {
  Handler *handlers = (Handler *) CF_ALLOC(sizeof(Handler) * count);
  Handler *h;
  for (h = handlers; h < handlers + count; h++) {
    h->class = cl;
    h->start_pc = readu2(cf);
    h->end_pc = readu2(cf);
    h->handler_pc = readu2(cf);
    h->catch_type = readu2(cf);
    h->data = 0;
  }
  return handlers;
}

static Method *read_methods(FILE *cf,Class *cl,int count) {
  Method *methods = (Method *) CF_ALLOC(sizeof(Method) * count);
  Method *m;
  u2 acount;
  int i;
  for (m = methods; m < methods + count; m++) {
    m->class = cl;
    m->access_flags = readu2(cf);
    m->name_index = readu2(cf);
    m->descriptor_index = readu2(cf);
    acount = readu2(cf); /* attribute count */
    while (acount-- > 0) {
      u2 index = readu2(cf);
      char *name = get_utf8(cl,index);
      if (STR_EQ(name,"Code")) {
	readu4(cf); /* skip attribute length */
	m->max_stack = readu2(cf);
	m->max_locals = readu2(cf);
	m->code_length = readu4(cf); 
	m->code = (u1 *) (CF_ALLOC(sizeof(u1) * m->code_length));
	for (i = 0; i < m->code_length; i++) 
	  m->code[i] = readu1(cf);
	m->exception_table_length = readu2(cf);
	m->exception_table = read_handlers(cf,cl,m->exception_table_length);
	skip_attributes(cf);
      } else if (STR_EQ(name,"Exceptions")) {
	readu4(cf); /* skip attribute length */
	m->number_of_exceptions = readu2(cf);
	for (i = 0; i < m->number_of_exceptions; i++)
	  m->exception_index_table[i] = readu2(cf);
      } else
	skip_attribute(cf);
    }
    m->data = 0;
  }
  return methods;
}

static Field *read_fields(FILE *cf,Class *cl,int count) {
  Field *fields = (Field *) CF_ALLOC(sizeof(Field) * count);
  Field *f;
  u2 acount;
  for (f = fields; f < fields + count; f++) {
    f->class = cl;
    f->access_flags = readu2(cf);
    f->name_index = readu2(cf);
    f->descriptor_index = readu2(cf);
    f->constantvalue_index = 0;  /* default if not present */
    acount = readu2(cf); /* attribute count */
    while (acount-- > 0) {
      u2 index = readu2(cf);
      char *name = get_utf8(cl,index);
      if (STR_EQ(name,"ConstantValue")) {
	readu4(cf); /* skip attribute length */
	f->constantvalue_index = readu2(cf);
      } else
	skip_attribute(cf);
    }
    f->data = 0;
  }
  return fields;
}

static Class *read_class(FILE *cf) {
  int i;
  Class *cl = (Class *) CF_ALLOC(sizeof(*cl)); 
  cl->magic = readu4(cf);
  cl->minor_version = readu2(cf);
  cl->major_version = readu2(cf);
  cl->constant_pool_count = readu2(cf); 
  cl->constant_pool = read_constant_pool(cf,cl->constant_pool_count);
  cl->access_flags = readu2(cf);
  cl->this_class = readu2(cf); 
  cl->super_class = readu2(cf);
  cl->interfaces_count = readu2(cf);
  cl->interfaces = (u2 *) CF_ALLOC(sizeof(u2) * cl->interfaces_count);
  for (i = 0; i < cl->interfaces_count; i++)
    cl->interfaces[i] = readu2(cf);
  cl->fields_count = readu2(cf);
  cl->fields = read_fields(cf,cl,cl->fields_count);
  cl->methods_count = readu2(cf);
  cl->methods = read_methods(cf,cl,cl->methods_count);
  skip_attributes(cf);
  cl->data = 0;
  return cl;
}

Class *load_class(char *name) {
  Class *class;
  char file[MAX_NAME+6];
  FILE *cf;
  strcpy(file,name);
  strcat(file,".class");
  cf = fopen(file,"r");
  if (!cf) 
    die ("cannot load class %s : %s",name,strerror(errno));
  class = read_class(cf);
  fclose(cf);
  return class;
}

char *class_as_type(char *class_name) {
  char *type = (char *) CF_ALLOC(strlen(class_name) + 3);
  char *t = type;
  *t++ = 'L';
  while (*class_name) 
    *t++ = *(class_name++);
  *t++ = ';';
  *t = '\0';
  return type;
}

