/* class.h
   Support for loading and querying class file data structure. 
   Handles all class files, but ignores some attibutes (including all
   non-standard ones).
*/

/* Handy class file type synonyms */
typedef uint8_t	        u1;
typedef uint16_t	u2;
typedef uint32_t	u4;

/* Access flag bit values. */
#define ACC_PUBLIC       0x0001  /* class, interface, field, method */
#define ACC_PRIVATE      0x0002  /* class field, class/instance method */
#define ACC_PROTECTED    0x0004  /* class field, class/instance method  */
#define ACC_STATIC       0x0008  /* field, class/instance method */
#define ACC_FINAL        0x0010  /* class, field, class/instance method */
#define ACC_SUPER        0x0020  /* class, interface  */
#define ACC_SYNCHRONIZED 0x0020  /* class/instance method */
#define ACC_VOLATILE     0x0040  /* class field */
#define ACC_TRANSIENT    0x0080  /* class field */
#define ACC_NATIVE       0x0100  /* class/instance method */
#define ACC_INTERFACE    0x0200  /* interface */
#define ACC_ABSTRACT     0x0400  /* class, interface, method */

/* Constant pool tag definitions. */
#define	CONSTANT_Class			7
#define	CONSTANT_Fieldref		9
#define	CONSTANT_Methodref		10
#define	CONSTANT_InterfaceMethodref	11
#define	CONSTANT_String			8
#define	CONSTANT_Integer		3
#define	CONSTANT_Float			4
#define	CONSTANT_Long			5
#define	CONSTANT_Double			6
#define	CONSTANT_NameAndType		12
#define	CONSTANT_Utf8			1

/* Array types */
#define T_BOOLEAN 4
#define T_CHAR    5
#define T_FLOAT   6
#define T_DOUBLE  7
#define T_BYTE    8
#define T_SHORT   9
#define T_INT    10
#define T_LONG   11

/* Constant pool entries */
typedef struct {
  u1 tag;
  union {
    struct {
      u2 name_index;
    } Class;
    struct {
      u2 class_index;
      u2 name_and_type_index;
    } Fieldref;
    struct {
      u2 class_index;
      u2 name_and_type_index;
    } Methodref;
    struct {
      u2 class_index;
      u2 name_and_type_index;
    } InterfaceMethodref;
    struct {
      u2 string_index;
    } String;
    struct {
      u4 bytes;
    } Integer;
    struct {
      u4 bytes;
    } Float;
    struct {
      u4 high_bytes;
      u4 low_bytes;
    } Long;
    struct {
      u4 high_bytes;
      u4 low_bytes;
    } Double;
    struct {
      u2 name_index;
      u2 descriptor_index;
    } NameAndType;
    struct {
      u2 length;
      u1 *bytes; /* null terminated, terminator not in length */
    } Utf8;
  } u;
} Constant;

typedef struct HandlerStruct Handler;
typedef struct MethodStruct Method;
typedef struct FieldStruct Field;
typedef struct ClassStruct Class;

/* Exception Handlers */
struct HandlerStruct {
  Class *class;
  u2 start_pc;
  u2 end_pc;
  u2 handler_pc;
  u2 catch_type;
  void *data;                  /* client-defined data */
};


/* Methods */
struct MethodStruct {
  Class *class;
  u2 access_flags;
  u2 name_index;
  u2 descriptor_index;
  /* from Code attribute: */
  u2 max_stack;   
  u2 max_locals;
  u4 code_length;
  u1 *code;                        /* indexed from 0 to code_length-1 */
  u2 exception_table_length;
  Handler *exception_table;        /* indexed from 0 to exception_table_length-1 */
  /* from Exceptions attribute: */
  u2 number_of_exceptions;
  u2 *exception_index_table;       /* indexed from 0 to exception_index_table-1 */
  void *data;                      /* client-defined data */
};

/* Fields */
struct FieldStruct {
  Class *class;
  u2 access_flags;
  u2 name_index;
  u2 descriptor_index;
  u2 constantvalue_index;  /* from ConstantValue attribute; 0 if none */
  void *data;              /* client-defined data */
};

/* Classes */
struct ClassStruct {
  u4 magic;
  u2 minor_version;
  u2 major_version;
  u2 constant_pool_count; 
  Constant *constant_pool; /* NB: indexed from 1 to constant_pool_count-1 */
  u2 access_flags;
  u2 this_class;
  u2 super_class;
  u2 interfaces_count;
  u2 *interfaces;          /* indexed from 0 to interfaces_count-1 */
  u2 fields_count;        
  Field *fields;           /* indexed from 0 to fields_count-1 */
  u2 methods_count;
  Method *methods;         /* indexed from 0 to methods_count-1 */
  /* no attributes for now */
  void *data;              /* client-defined data */
};

/* Constant Pool Access functions.  
   These should always be used in preference to poking around
   directly in the pool.
 */

extern u1   get_constant_tag(Class *class, u2 index);
extern char *get_utf8(Class *class,u2 index);
extern char *get_class(Class *class,u2 index);
extern char *get_string(Class *class,u2 index);
extern int32_t get_integer(Class *class,u2 index);
extern char *get_name(Class *class,u2 index);
extern char *get_type(Class *class,u2 index);
extern char *get_field_class(Class *class,u2 index);
extern char *get_field_name(Class *class,u2 index);
extern char *get_field_type(Class *class,u2 index);
extern char *get_method_class(Class *class,u2 index);
extern char *get_method_name(Class *class,u2 index);
extern char *get_method_type(Class *class,u2 index);
extern char *get_imethod_class(Class *class,u2 index);
extern char *get_imethod_name(Class *class,u2 index);
extern char *get_imethod_type(Class *class,u2 index);

/* Find a method in a class's method table. */
extern Method *find_method(Class *class,char *method_name, char *method_type);

/* Return size (in words) of value represented by descriptor. 
   (Only first character of descriptor is needed.) */
size_t size_descriptor(char *s);

/* Return the number of words occupied by method's arguments. */
extern size_t count_args(char *method_type);

/* Return the number of words occupied by method's result. */
extern size_t count_result(char *method_type);

#define MAX_NAME 65536

/* Load a class given its file name. */
extern Class *load_class(char *name);

/* Return a type descriptor for a reference to the named class */
extern char *class_as_type(char *class_name);

#define OBJECT "java/lang/Object"

