/* Initial interpreter for extended JVM subset.

   Changes from subset 1:
   Supports NEW, INVOKESPECIAL, GETFIELD, PUTFIELD on arbitrary classes 
   (but these must be in current working directory). 
   Optional second command-line parameter specifies heap size (in words).
*/


#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdint.h>
#include "basics.h"
#include "class.h"
#include "objects.h"
#include "bytecode.h"
#include "runtime.h"

#define HEAPSIZE 32768  /* words */

#define STACKSIZE 32768 /* words */

/* Globals */

/* N.B. The stack grows up */
static uintptr_t *stack; /* pointer to base of stack */
static uintptr_t *limit; /* pointer to one past last stack slot */
static uintptr_t *sp;    /* pointer to highest occupied stack slot */ 


/* The type stack (moves in parallel with stack): contains 
   type desriptors for each slot. */
char **typestack;   /* same size as stack */
char **tsp;         /* typestack pointer */

/* Loaded class cache. */
#define MAX_CLASSES 100
static Class *classes[MAX_CLASSES];
static int class_count = 0;

/* Locate a class in the cache, loading it if not yet present. */
static Class *find_class(char *name) {
  int i;
  for (i = 0; i < class_count; i++) 
    if (STR_EQ(get_class(classes[i],classes[i]->this_class),name))
      return classes[i];

  /* class not yet present: get it */
  if (class_count < MAX_CLASSES) 
    return (classes[class_count++] = load_class(name));
  else {
    die("too many classes");
    return 0;  // quiet compiler warning
  }
}

/* ---------------------------------------------------------- */

/* Interpret a method. */
static void interp(Class *class,Method *method) {
  u1 *pc = method->code;  /* program counter */
  size_t nargs = count_args(get_utf8(class,method->descriptor_index)) + ((method->access_flags & ACC_STATIC) == 0 ? 1 : 0);
  uintptr_t *locals = sp - nargs + 1; /* address of local #0 */
  char **tlocals = tsp - nargs + 1; /* address of local #0 type desc */
  sp = locals + method->max_locals - 1;  /* allocate space for locals */
  while (tsp < tlocals + method->max_locals - 1) /* initialize local type descs */
    *(++tsp) = "I";     /* safe descriptor for uninitialized values */
  if (sp + method->max_stack >= limit) /* check there is room for operand stack. */
    exception(STACK_OVERFLOW_EXN);
  while (1) {
    switch (*pc) {
    case ACONST_NULL:
      {  
	*(++sp) = (uintptr_t) 0;
	*(++tsp) = class_as_type(OBJECT);
	pc++;
	break;
      }
    case ARETURN:
      { 
	uintptr_t *v = (uintptr_t *) *sp;
	char *t = *tsp;
	sp = locals;
	tsp = tlocals;
	*sp = (uintptr_t) v;
	*tsp = t;
	return;
      }
    case ARRAYLENGTH:
      {
	uintptr_t *a = (uintptr_t *) (*sp);
 	if (a == 0)
 	  exception (NULL_PTR_EXN);
 	*sp = (uintptr_t) (array_length(a));
	*tsp = "I";
 	pc++;
	break;
      }
    case BIPUSH:
      {	
	*(++sp) = (uintptr_t)((int8_t)(pc[1]));
	*(++tsp) = "I";
        pc+=2;
	break; 
      }
    case DUP: 
      {	
	sp++;
	tsp++;
	*sp = *(sp-1);
	*tsp = *(tsp-1);
	pc++;
	break;
      }
    case DUP2:
      {
	sp+=2;
	tsp+=2;
	*sp = *(sp-2);
	*(sp-1) = *(sp-3);
	*tsp = *(tsp-2);
	*(tsp-1) = *(tsp-3);
	pc++;
	break;
      }
    case GETFIELD:
      {
	u2 index = (pc[1] << 8) | pc[2];
	char *class_name = get_field_class(class,index);
	char *field_name = get_field_name(class,index);
	char *field_type = get_field_type(class,index);
	Class *target_class = find_class(class_name);
	int offset = object_field_offset(target_class,field_name,field_type);
	uintptr_t *object = (uintptr_t *) (*sp);
	if (object == 0)
	  exception (NULL_PTR_EXN);
	*sp = get_field(object,offset);
	*tsp = field_type;
	pc+=3;
	break;
      }
    case GETSTATIC:
      { 
	u2 index = (pc[1] << 8) | pc[2];
	/* kludge for java.lang.System.out */
	if (STR_EQ(get_field_class(class,index),"java/lang/System") &&
	    STR_EQ(get_field_name(class,index),"out")) {
	  sp++;  /* just push an empty slot */
	  *(++tsp) = "I";
	} else 
	  die("unimplemented instruction GETSTATIC");
	pc+=3;
	break;
      }
    case GOTO:
      {
	int16_t offset = (pc[1] << 8) | pc[2];
	pc += offset;
	break;
      }


#define IBINOP(op) \
      { \
	int32_t v2 = (int32_t) (*(sp--)); \
	int32_t v1 = (int32_t) (*sp);\
	*sp = (uintptr_t) (v1 op v2);\
        tsp--;\
	pc++;\
	break;\
      }

    case IADD:
      IBINOP(+);
    case IALOAD:
      {
 	int32_t index = (int32_t) (*(sp--));
 	uintptr_t *a = (uintptr_t *) (*sp);
 	int32_t length;
 	if (a == 0)
 	  exception (NULL_PTR_EXN);
 	length = array_length(a);
 	if (index < 0 || index >= length)
 	  exception (INDEX_BOUNDS_EXN);
 	*sp = (uintptr_t) (array_load(a,index));
	*(--tsp) = "I";
 	pc ++;
 	break;
       }
    case IAND:
      IBINOP(&);
    case IASTORE:
      {
	int32_t v = (int32_t) (*(sp--));
	int32_t index = (int32_t) (*(sp--));
	uintptr_t *a = (uintptr_t *) (*sp--);
	int32_t length;
	if (a == 0)
	  exception (NULL_PTR_EXN);
	length = array_length(a);
	if (index < 0 || index >= length)
	  exception (INDEX_BOUNDS_EXN);
	array_store(a,index,v);
	tsp -= 3;
	pc ++;
	break;
      }

#define ICONST(v)\
      {  *(++sp) = (uintptr_t) ((int32_t) (v));	\
         *(++tsp) = "I";\
         pc++;\
         break;\
      }

    case ICONST_M1:
      ICONST(-1);
    case ICONST_0:
      ICONST(0);
    case ICONST_1:
      ICONST(1);
    case ICONST_2:
      ICONST(2);
    case ICONST_3:
      ICONST(3);
    case ICONST_4:
      ICONST(4);
    case ICONST_5:
      ICONST(5);

#define IF_ACMP(op) \
      { \
	uintptr_t *a2 = (uintptr_t *) (*(sp--)); \
	uintptr_t *a1 = (uintptr_t *) (*(sp--)); \
        tsp -=2; \
	int16_t offset = (pc[1] << 8) | pc[2]; \
	if (a1 op a2) \
	  pc += offset; \
	else \
	  pc += 3; \
	break; \
     }

    case IF_ACMPEQ:
      IF_ACMP(==);
    case IF_ACMPNE:
      IF_ACMP(!=);

#define IF_ICMP(op) \
      { \
	int32_t v2 = (int32_t) (*(sp--)); \
	int32_t v1 = (int32_t) (*(sp--)); \
        tsp -=2; \
	int16_t offset = (pc[1] << 8) | pc[2]; \
	if (v1 op v2) \
	  pc += offset; \
	else \
	  pc += 3; \
	break; \
      }

    case IF_ICMPEQ:
      IF_ICMP(==);
    case IF_ICMPGE:
      IF_ICMP(>=);
    case IF_ICMPGT:
      IF_ICMP(>);
    case IF_ICMPLE:
      IF_ICMP(<=);
    case IF_ICMPLT:
      IF_ICMP(<);
    case IF_ICMPNE:
      IF_ICMP(!=);

#define IF0(op) \
      {\
	int32_t v = (int32_t) (*(sp--));\
        tsp--;\
	int16_t offset = (pc[1] << 8) | pc[2];\
	if (v op 0)\
	  pc += offset;\
	else\
	  pc += 3;\
	break;\
      }

    case IFEQ:
      IF0(==);
    case IFGE:
      IF0(>=);
    case IFGT:
      IF0(>);
    case IFLE:
      IF0(<=);
    case IFLT:
      IF0(<);
    case IFNE:
      IF0(!=);

#define IFA0(op) \
      {\
	uintptr_t *a = (uintptr_t *) (*(sp--));\
	tsp--;\
	int16_t offset = (pc[1] << 8) | pc[2];\
	if (a op 0)\
	  pc += offset;\
	else\
	  pc += 3;\
	break;\
      }

    case IFNONNULL:
      IFA0(!=);
    case IFNULL:
      IFA0(==);

    case IINC:
      {
	u1 index = pc[1];
	int8_t incr = (int8_t) (pc[2]);
	locals[index] += incr;
	pc += 3;
	break;
      }
    case ALOAD:
    case ILOAD:
      {
	u1 index = pc[1];
	*(++sp) = locals[index];
	*(++tsp) = tlocals[index];
	pc += 2;
	break;
      }

#define ILOADL(loc)\
      {\
	*(++sp) = locals[loc];\
        *(++tsp) = tlocals[loc];\
	pc ++;\
	break;\
      }

    case ALOAD_0:
    case ILOAD_0:
      ILOADL(0);
    case ALOAD_1:
    case ILOAD_1:
      ILOADL(1);
    case ALOAD_2:
    case ILOAD_2:
      ILOADL(2);
    case ALOAD_3:
    case ILOAD_3:
      ILOADL(3);

    case IMUL:
      IBINOP(*);

    case INEG:
      {
	int32_t v = (int32_t) (*sp);
	*sp = (uintptr_t) (-v);
	pc++;
	break;
      }
    case INVOKESPECIAL:
      {
	u2 index = (pc[1] << 8) | pc[2];
	char *target_class_name = get_method_class(class,index);
	char *target_method_name = get_method_name(class,index);
	char *target_method_type = get_method_type(class,index);
	if (!STR_EQ(target_method_name,"<init>"))
	  die ("INVOKESPECIAL only supported on <init> methods"); 
	if (!STR_EQ(target_class_name,OBJECT)) {
	  Class *target_class = find_class(target_class_name);
	  size_t target_nargs = count_args(target_method_type) + 1; 
	  uintptr_t *object = (uintptr_t*) (*(sp-(target_nargs-1)));
	  if (object == 0)
	    exception (NULL_PTR_EXN);
	  Method *target_method = find_method(target_class,
					      target_method_name,
					      target_method_type);
	  interp(target_class,target_method);
	} else {  // Object.<init> has no method to execute
	  sp--;   // just pop (sole) argument       
	  tsp--;
	}
	pc+=3;
	break;
      }
    case INVOKESTATIC:
      {
	u2 index = (pc[1] << 8) | pc[2];
	char *target_class = get_method_class(class,index);
	/* only support static calls into our own class */
	if (STR_EQ(target_class,get_class(class,class->this_class))) {
	  Method *m = find_method(class,
			 	  get_method_name(class,index),
				  get_method_type(class,index));
	  interp(class,m);
	} else
	  die ("unimplemented INVOKESTATIC to method in different class %s", target_class);
	pc +=3;
	break;
      }
    case INVOKEVIRTUAL:
      { 
	u2 index = (pc[1] << 8) | pc[2];
	/* kludge for java.io.PrintStream.print */
	if (STR_EQ(get_method_class(class,index),"java/io/PrintStream")
	    && STR_EQ(get_method_name(class,index),"print")) {
	  char *type = get_method_type(class,index);
	  if (STR_EQ(type,"(Ljava/lang/String;)V")) {
	    /* arg is string */
	    print_string((char *)(*(sp--)));
	  } else if(STR_EQ(type,"(I)V")) {
	    /* arg is int */
	    print_int((int32_t)(*(sp--)));
	  } else
	    die ("unimplemented PrintStream.print method with signature %s", type);
	  sp--;  /* pop the empty slot corresponding to the object argument */
	  tsp -= 2;
	} else 
	  die("unimplemented instruction INVOKEVIRTUAL");
	pc+=3;
	break;
      }
    case IOR:
      IBINOP(|);

    case IRETURN:
      { 
	int32_t v = (int32_t) (*sp);
	char *t = *tsp;
	sp = locals;
	tsp = tlocals;
	*sp = (uintptr_t) v;
	*tsp = t;
	return;
      }
    case ASTORE:
    case ISTORE:
      {
	u1 index = pc[1];
	locals[index] = *(sp--);
	tlocals[index] = *(tsp--);
	pc += 2;
	break;
      }
#define ISTOREL(loc)\
      {\
	locals[loc] = *(sp--);\
        tlocals[loc] = *(tsp--);\
	pc++;\
	break;\
      }

    case ASTORE_0:
    case ISTORE_0:
      ISTOREL(0);
    case ASTORE_1:
    case ISTORE_1:
      ISTOREL(1);
    case ASTORE_2:
    case ISTORE_2:
      ISTOREL(2);
    case ASTORE_3:
    case ISTORE_3:
      ISTOREL(3);
    case ISUB:
      IBINOP(-);
    case IXOR:
      IBINOP(^);
    case LDC:
      {
	u1 index = pc[1];
	u1 tag = get_constant_tag(class,index);
	if (tag == CONSTANT_String) 
	  *(++sp) = (uintptr_t) (get_string(class,index));
	else if (tag == CONSTANT_Integer) 
	  *(++sp) = (uintptr_t) (get_integer(class,index));
	else 
	  die ("unsupported constant type %d", tag);
	*(++tsp) = "I";  // close enough
	pc+=2;
	break;
      }
    case LDC_W:
      {
	u2 index = (pc[1] << 8) | pc[2];
	u1 tag = get_constant_tag(class,index);
	if (tag == CONSTANT_String) 
	  *(++sp) = (uintptr_t) (get_string(class,index));
	else if (tag == CONSTANT_Integer) 
	  *(++sp) = (uintptr_t) (get_integer(class,index));
	else 
	  die ("unsupported constant type %d", tag);
	*(++tsp) = "I";  // close enough
	pc+=3;
	break;
      }
    case NEW:
      {
	u2 index = (pc[1] << 8) | pc[2];
	char *class_name = get_class(class,index);
	Class *target_class = find_class(class_name); 
	uintptr_t obj = (uintptr_t) alloc_object(target_class);  // be careful to do this *before* altering sp !
	*(++sp) = obj;
	*(++tsp) = class_as_type(class_name);
	pc+=3;
	break;
      }
    case NEWARRAY:
      {
	u1 atype = pc[1];
	int32_t count;
	int32_t *a;
	int i;
	if (atype != T_INT) 
	  die("unimplemented array type %d",atype);
	count = (int32_t) (*sp);
	*sp = (uintptr_t) alloc_array(count);
	*tsp = "[I";
	pc += 2;
	break;
      }
    case NOP: 
      {
	pc++;
	break;
      }
    case POP:
      {
	sp--;
	tsp--;
	pc++;
	break;
      }
    case POP2:
      {
	sp -= 2;
	tsp -= 2;
	pc++;
	break;
      }
    case PUTFIELD:
      {
	u2 index = (pc[1] << 8) | pc[2];
	char *class_name = get_field_class(class,index);
	char *field_name = get_field_name(class,index);
	char *field_type = get_field_type(class,index);
	Class *target_class = find_class(class_name);
	int offset = object_field_offset(target_class,field_name,field_type);
	uintptr_t value = *sp--;
	uintptr_t *object = (uintptr_t *) (*sp--);
	if (object == 0)
	  exception (NULL_PTR_EXN);
	put_field(object,offset,value);
	tsp -= 2;
	pc+=3;
	break;
      }
    case RETURN:
      {
	sp = locals - 1;
	tsp = tlocals - 1;
	return;
      }
    case SIPUSH:
      {
	*(++sp) = (uintptr_t) ((int16_t) ((pc[1] << 8) | pc[2]));
	*(++tsp) = "I";
	pc +=3;
	break;
      }
    case SWAP:
      {
	uintptr_t v = *sp;
	char *t = *tsp;
	*sp = *(sp-1);
	*tsp = *(tsp-1);
	*(sp-1) = v;
	*(tsp-1) = t;
	pc++;
	break;
      }
    default: 
      die("unimplemented instruction code %i at offset %d", *pc, pc - method->code);
    }
  }
  /* supposedly can't fall off the end */
  assert(false);
}

int main(int argc, char **argv) {
  Class *class = load_class(*++argv);        
  int heap_size = (argc > 2) ? atoi (*++argv) : HEAPSIZE;

  Method *main_method = find_method(class,"main","([Ljava/lang/String;)V");
  if (!main_method) 
    die ("cannot find appropriate 'main' method in class\n");
  

  /* initialize stacks */
  stack = ((uintptr_t *) alloc(sizeof(uintptr_t) * STACKSIZE));
  sp = stack - 1;
  limit = stack + STACKSIZE;
  typestack = ((char **) alloc(sizeof(char *) * STACKSIZE));
  tsp = typestack - 1;
  
  initialize_heap(heap_size,stack,&sp,typestack);

  /* push dummy argv parameter for main */
  *(++sp) = 0;
  *(++tsp) = "[Ljava/lang/String;";

  interp(class,main_method);
  uintptr_t ret = *sp;
  return (int) ret;
}
