#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <stdint.h>
#include "basics.h"
#include "class.h"
#include "objects.h"
#include "runtime.h"


/* Exceptions */

/* Warning: must match order in enum declaration for exn */
static char *exn_name[] = 
  {"OutOfMemory",
   "StackOverflow",
   "NullPointer",
   "ArrayIndexOutOfBounds",
   "NegativeArraySize"};

void exception(exn reason) {
  fprintf(stderr,"Uncaught Exception: %s\n",exn_name[reason]);
  exit(1);
}

/* Printing */

void print_int(int i) {
  printf("%d",i);
}

void print_string(char *s) {
  printf("%s",s);
}

/* Heap allocation */
/* all sizes are in uintptr_t sized words */

typedef struct space {          /* allocation space data              */
  uintptr_t *start;             /* first word of space                */
  uintptr_t *end;               /* one past last word of space        */
  uintptr_t *avail;             /* pointer to next word into which to create an object  */
} space;

/* allocate and initialize descriptor structure and storage for a space */
space *initspace(size_t words) {
  space *s = alloc(sizeof(space));
  s->start = alloc(words*sizeof(uintptr_t));
  s->end = s->start + words;
  s->avail = s->start;
  return s;
}

space *heap;

/* initialize the heap 
 * heap_size is total desired size (in words) 
 * first_root points to first entry in array of roots
 * last_root_ptr points to pointer to last entry in array of roots
 * root_types points to first entry in parallel array of type descriptors (same length as root array)
 */
void initialize_heap(size_t heap_words,uintptr_t *first_root,uintptr_t **last_root_ptr,char **root_types) {
  heap = initspace(heap_words);
  /* for now, ignore root information */
}
  
/* allocate a heap record of specified number of words */
uintptr_t *heapalloc(size_t words) {
  uintptr_t *p;
  if (heap->end - heap->avail < words)
    exception(OUT_OF_MEMORY_EXN);
  p = heap->avail;
  heap->avail += words;
  return p;
}

/* Interpreter operations */

/* Array operations */

uintptr_t *alloc_array(int32_t size) {
  if (size < 0)
    exception(ARRAY_SIZE_EXN);
  uintptr_t *a = heapalloc(size+1) + 1;
  a[-1] = size;
  int i;
  for (i = 0; i < size; i++)
    a[i] = 0;
  return a;
}

int32_t array_length(uintptr_t *a) {
  return (int32_t) (*(a-1));
}

void array_store(uintptr_t *a,int32_t index,int32_t v) {
  *(a+index) = (uintptr_t) v;
}

int32_t array_load(uintptr_t *a,int32_t index) {
  return (int32_t) (*(a + index));
}

/* Object and field operations */

uintptr_t *alloc_object(Class *class) {
  size_t size = count_object_slots(class);
  uintptr_t *obj = heapalloc(size);
  int i;
  for (i = 0; i < size; i++)
    obj[i] = 0;
  return obj;
}

uintptr_t get_field(uintptr_t *object,size_t offset) {
  return *(object+offset);
}

void put_field(uintptr_t *object,size_t offset,uintptr_t value) {
  *(object + offset) = value;
}


